import argparse
import sys
from importlib.metadata import version
from pathlib import Path

from bolthole.git import GitRepo, configure_output
from bolthole.watcher import watch


def main():
    parser = argparse.ArgumentParser(prog="bolthole")
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version v{version('bolthole')}",
    )
    parser.add_argument(
        "--watchdog-debug",
        action="store_true",
        help="show raw filesystem events",
    )
    parser.add_argument(
        "-n",
        "--dry-run",
        action="store_true",
        help="take no actions, but report what would happen",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="show file updates as well as actions taken",
    )
    parser.add_argument(
        "--timeless",
        action="store_true",
        help="omit timestamps from output",
    )
    parser.add_argument(
        "--ignore",
        action="append",
        default=[],
        metavar="PATTERN",
        help="ignore files matching pattern (repeatable)",
    )
    parser.add_argument(
        "--show-git",
        action="store_true",
        help="display git commands and their output",
    )
    parser.add_argument(
        "--once",
        action="store_true",
        help="commit and exit without watching for changes",
    )
    parser.add_argument(
        "-a",
        "--author",
        metavar="AUTHOR",
        help="override commit author (format: 'Name <email>')",
    )
    parser.add_argument(
        "-b",
        "--bundle",
        type=float,
        default=0,
        metavar="SECONDS",
        help="bundle files older than threshold into single commit",
    )
    parser.add_argument(
        "-g",
        "--grace",
        type=float,
        default=0,
        metavar="SECONDS",
        help="delay commits by grace period (default: 0)",
    )
    parser.add_argument(
        "-m",
        "--message",
        metavar="MESSAGE",
        help="override commit message",
    )
    parser.add_argument(
        "-r",
        "--remote",
        action="append",
        default=[],
        metavar="REMOTE",
        help="push to remote after commit (repeatable)",
    )
    parser.add_argument("source")
    parser.add_argument("dest", nargs="?")
    args = parser.parse_args()

    configure_output(args.timeless)

    if args.grace < 0:
        print("error: grace period cannot be negative", file=sys.stderr)
        sys.exit(2)

    if args.bundle < 0:
        print("error: bundle threshold cannot be negative", file=sys.stderr)
        sys.exit(2)

    if args.bundle > 0 and args.grace == 0:
        print("error: bundle requires grace period", file=sys.stderr)
        sys.exit(2)

    if args.grace and args.bundle >= args.grace:
        print("error: bundle threshold must be less than grace period", file=sys.stderr)
        sys.exit(2)

    source = Path(args.source).resolve()
    if not source.exists():
        print(
            f"error: source directory does not exist: {args.source}",
            file=sys.stderr,
        )
        sys.exit(2)

    dest = None
    if not args.dest:
        if not GitRepo.is_repo(source):
            print("error: source must be a git repository in single-directory mode",
                  file=sys.stderr)
            sys.exit(2)
    else:
        dest = Path(args.dest).resolve()
        if source == dest:
            print("error: source and destination cannot be the same",
                  file=sys.stderr)
            sys.exit(2)
        if source.is_relative_to(dest):
            print("error: source cannot be inside destination",
                  file=sys.stderr)
            sys.exit(2)
        if dest.is_relative_to(source):
            print("error: destination cannot be inside source",
                  file=sys.stderr)
            sys.exit(2)

        if GitRepo.is_repo(dest):
            pass
        elif dest.exists() and any(dest.iterdir()):
            print("error: destination exists but is not a git repository",
                  file=sys.stderr)
            sys.exit(2)
        elif not args.dry_run:
            dest.mkdir(parents=True, exist_ok=True)
            repo = GitRepo(dest)
            repo.init()

    if args.remote:
        if dest:
            repo_path = dest
        else:
            repo_path = source
        repo = GitRepo(repo_path)
        for remote in args.remote:
            if not repo.has_remote(remote):
                print(f"remote '{remote}' needs to be added",
                      file=sys.stderr)
                sys.exit(2)

    watch(
        source,
        dest=dest,
        dry_run=args.dry_run,
        verbose=args.verbose,
        watchdog_debug=args.watchdog_debug,
        ignore_patterns=args.ignore,
        show_git=args.show_git,
        source_label=args.source,
        dest_label=args.dest,
        once=args.once,
        author=args.author,
        message=args.message,
        remotes=args.remote,
        grace=args.grace,
        bundle=args.bundle,
    )



from textwrap import dedent

from bolthole.debounce import Event
from bolthole.git import GitRepo


class TestEmpty:
    def test_empty_returns_empty_string(self):
        assert GitRepo.generate_commit_message([]) == ""


class TestAdd:
    def test_one_add(self):
        assert GitRepo.generate_commit_message([
            Event("created", "café.txt"),
        ]) == "Add café.txt"

    def test_two_adds(self):
        assert GitRepo.generate_commit_message([
            Event("created", "one.txt"),
            Event("created", "two.txt"),
        ]) == "Add one.txt and two.txt"

    def test_three_adds(self):
        assert GitRepo.generate_commit_message([
            Event("created", "one.txt"),
            Event("created", "three.txt"),
            Event("created", "two.txt"),
        ]) == "Add one.txt, three.txt, and two.txt"

    def test_long_filename(self):
        assert GitRepo.generate_commit_message([
            Event("created", "project/alpha/components/widgets/buttons/file.txt"),
        ]) == dedent("""
            Add 1 file

            - Add project/alpha/components/widgets/buttons/file.txt
        """).strip()

    def test_two_long_filenames(self):
        assert GitRepo.generate_commit_message([
            Event("created", "project/alpha/components/widgets/button.txt"),
            Event("created", "project/alpha/components/widgets/input.txt"),
        ]) == dedent("""
            Add 2 files

            - Add project/alpha/components/widgets/button.txt
            - Add project/alpha/components/widgets/input.txt
        """).strip()

    def test_short_and_long_filename(self):
        assert GitRepo.generate_commit_message([
            Event("created", "file.txt"),
            Event("created", "project/alpha/components/widgets/button.txt"),
        ]) == dedent("""
            Add 2 files

            - Add file.txt
            - Add project/alpha/components/widgets/button.txt
        """).strip()


class TestUpdate:
    def test_one_update(self):
        assert GitRepo.generate_commit_message([
            Event("modified", "file.txt"),
        ]) == "Update file.txt"

    def test_two_updates(self):
        assert GitRepo.generate_commit_message([
            Event("modified", "my file.txt"),
            Event("modified", "your file.txt"),
        ]) == "Update my file.txt and your file.txt"

    def test_three_updates(self):
        assert GitRepo.generate_commit_message([
            Event("modified", "one.txt"),
            Event("modified", "three.txt"),
            Event("modified", "two.txt"),
        ]) == "Update one.txt, three.txt, and two.txt"


class TestRemove:
    def test_one_remove(self):
        assert GitRepo.generate_commit_message([
            Event("deleted", "file.txt"),
        ]) == "Remove file.txt"

    def test_two_removes(self):
        assert GitRepo.generate_commit_message([
            Event("deleted", "one.txt"),
            Event("deleted", "two.txt"),
        ]) == "Remove one.txt and two.txt"

    def test_three_removes(self):
        assert GitRepo.generate_commit_message([
            Event("deleted", "one.txt"),
            Event("deleted", "three.txt"),
            Event("deleted", "two.txt"),
        ]) == "Remove one.txt, three.txt, and two.txt"


class TestRename:
    def test_one_rename(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "naïve.txt", "naïveté.txt"),
        ]) == "Rename naïve.txt to naïveté.txt"

    def test_two_renames(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "old1.txt", "new1.txt"),
            Event("renamed", "old2.txt", "new2.txt"),
        ]) == dedent("""
            Rename 2 files

            - Rename old1.txt to new1.txt
            - Rename old2.txt to new2.txt
        """).strip()

    def test_three_renames(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "old1.txt", "new1.txt"),
            Event("renamed", "old2.txt", "new2.txt"),
            Event("renamed", "old3.txt", "new3.txt"),
        ]) == dedent("""
            Rename 3 files

            - Rename old1.txt to new1.txt
            - Rename old2.txt to new2.txt
            - Rename old3.txt to new3.txt
        """).strip()

    def test_long_rename(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "project/old/path/file.txt", "project/new/path/file.txt"),
        ]) == dedent("""
            Rename 1 file

            - Rename project/old/path/file.txt to project/new/path/file.txt
        """).strip()


class TestCombinations:
    def test_one_of_each(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "old.txt", "new.txt"),
            Event("created", "added.txt"),
            Event("deleted", "removed.txt"),
            Event("modified", "updated.txt"),
        ]) == dedent("""
            Change 4 files

            - Add added.txt
            - Rename old.txt to new.txt
            - Remove removed.txt
            - Update updated.txt
        """).strip()

    def test_two_of_each(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "charlie.txt", "tango.txt"),
            Event("renamed", "golf.txt", "whiskey.txt"),
            Event("created", "alpha.txt"),
            Event("created", "echo.txt"),
            Event("deleted", "bravo.txt"),
            Event("deleted", "foxtrot.txt"),
            Event("modified", "delta.txt"),
            Event("modified", "hotel.txt"),
        ]) == dedent("""
            Change 8 files

            - Add alpha.txt
            - Remove bravo.txt
            - Rename charlie.txt to tango.txt
            - Update delta.txt
            - Add echo.txt
            - Remove foxtrot.txt
            - Rename golf.txt to whiskey.txt
            - Update hotel.txt
        """).strip()

    def test_add_and_update_short(self):
        assert GitRepo.generate_commit_message([
            Event("created", "added.txt"),
            Event("modified", "updated.txt"),
        ]) == "Add added.txt, update updated.txt"

    def test_rename_and_remove_short(self):
        assert GitRepo.generate_commit_message([
            Event("renamed", "old.txt", "new.txt"),
            Event("deleted", "gone.txt"),
        ]) == "Remove gone.txt, rename old.txt to new.txt"

    def test_long_filenames_mixed(self):
        assert GitRepo.generate_commit_message([
            Event("created", "project/source/components/button.tsx"),
            Event("modified", "project/source/components/input.tsx"),
        ]) == dedent("""
            Change 2 files

            - Add project/source/components/button.tsx
            - Update project/source/components/input.tsx
        """).strip()



import shlex
import subprocess
from datetime import datetime
from pathlib import Path

from bolthole.debounce import Event


_timeless = False


def configure_output(timeless: bool):
    global _timeless
    _timeless = timeless


def output(message: str):
    if _timeless:
        print(message, flush=True)
    else:
        ts = datetime.now().strftime("%H:%M:%S ")
        print(f"{ts}{message}", flush=True)


class GitRepo:
    SUBJECT_LINE_LIMIT = 50

    def __init__(
        self,
        path,
        dry_run=False,
        show_git=False,
        author=None,
        message=None,
    ):
        self.path = Path(path)
        self.dry_run = dry_run
        self.show_git = show_git
        self.author = author
        self.message = message

    EXCLUDE_FLAGS = {
        "--no-verify",
        "--no-gpg-sign",
        "--quiet",
    }
    EXCLUDE_OPTIONS = {
        "-m",
        "-C",
    }
    EXCLUDE_COMMANDS = {
        "diff",
        "status",
    }
    DRY_RUN_ALLOWED = {
        "diff",
        "status",
    }

    def run_git(self, *args, **kwargs):
        if args:
            command = args[0]
        else:
            command = None
        display_parts = ["git"]
        skip_next = False
        for arg in args:
            if skip_next:
                skip_next = False
                continue
            if arg in self.EXCLUDE_FLAGS:
                continue
            if arg in self.EXCLUDE_OPTIONS:
                skip_next = True
                continue
            display_parts.append(arg)

        formatted = " ".join(shlex.quote(arg) for arg in display_parts)

        if self.dry_run and command not in self.DRY_RUN_ALLOWED:
            output(f"#  {formatted}")
            return subprocess.CompletedProcess(args=[], returncode=0)

        show_this = self.show_git and command not in self.EXCLUDE_COMMANDS
        if show_this:
            output(f"%  {formatted}")

        result = subprocess.run(["git", "-C", str(self.path), *args], **kwargs)

        return result

    @staticmethod
    def is_repo(path):
        result = subprocess.run(
            ["git", "-C", str(path), "rev-parse", "--git-dir"],
            capture_output=True,
        )
        return result.returncode == 0

    def has_remote(self, name):
        result = subprocess.run(
            ["git", "-C", str(self.path), "remote", "get-url", name],
            capture_output=True,
        )
        return result.returncode == 0

    def push(self, remotes):
        for remote in remotes:
            result = self.run_git("push", remote, "HEAD", capture_output=True)
            if result.returncode != 0:
                output(f"!! push to {remote} failed")

    def init(self):
        self.run_git("init", "--quiet", check=True)

    def add_all(self):
        self.run_git("add", "-A", check=True)

    def get_uncommitted(self):
        result = self.run_git(
            "status", "--porcelain", "-z",
            capture_output=True, text=True, check=True,
        )
        events = []
        for entry in result.stdout.split("\0"):
            if not entry:
                continue
            status = entry[:2]
            path = entry[3:]
            if status == "??":
                events.append(Event("created", path))
            elif "D" in status:
                events.append(Event("deleted", path))
            elif "M" in status or "A" in status:
                events.append(Event("modified", path))
        return events

    def get_staged(self):
        result = self.run_git(
            "diff", "--cached", "--name-status",
            capture_output=True, text=True, check=True,
        )
        events = []
        for line in result.stdout.splitlines():
            parts = line.split("\t")
            status = parts[0]
            path = parts[1]
            if status == "A":
                events.append(Event("created", path))
            elif status == "D":
                events.append(Event("deleted", path))
            elif status == "M":
                events.append(Event("modified", path))
            elif status.startswith("R"):
                new_path = parts[2]
                events.append(Event("renamed", path, new_path))
        return events

    def commit_changes(self, events):
        if not events:
            return
        paths = []
        for event in events:
            paths.append(event.path)
            if event.new_path:
                paths.append(event.new_path)
        self.run_git("add", "--", *paths)
        if self.dry_run:
            if self.message:
                message = self.message
            else:
                message = self.generate_commit_message(events)
        else:
            staged = self.get_staged()
            if not staged:
                return
            if self.message:
                message = self.message
            else:
                message = self.generate_commit_message(staged)
        commit_args = [
            "commit",
            "-m", message,
            "--no-verify",
            "--no-gpg-sign",
            "--quiet",
        ]
        if self.author:
            commit_args.extend(["--author", self.author])
        self.run_git(*commit_args, check=True)

    @staticmethod
    def generate_commit_message(events):
        if not events:
            return ""

        verb_map = {
            "created": "Add",
            "modified": "Update",
            "deleted": "Remove",
            "renamed": "Rename",
        }

        items = []
        for event in events:
            verb = verb_map[event.type]
            if event.type == "renamed":
                line = f"{verb} {event.path} to {event.new_path}"
            else:
                line = f"{verb} {event.path}"
            items.append((event.path, line))

        items.sort(key=lambda x: x[0])
        count = len(events)

        types = set(e.type for e in events)
        if len(types) == 1:
            # only one operation, can simplify subject line
            event_type = next(iter(types))
            verb = verb_map[event_type]

            if event_type == "renamed":
                if count == 1:
                    candidate = items[0][1]
                    if len(candidate) < GitRepo.SUBJECT_LINE_LIMIT:
                        return candidate
            else:
                # oxford comma join filenames
                filenames = [item[0] for item in items]
                if len(filenames) == 1:
                    candidate = f"{verb} {filenames[0]}"
                elif len(filenames) == 2:
                    candidate = f"{verb} {filenames[0]} and {filenames[1]}"
                else:
                    joined = ', '.join(filenames[:-1])
                    candidate = f"{verb} {joined}, and {filenames[-1]}"

                if len(candidate) < GitRepo.SUBJECT_LINE_LIMIT:
                    return candidate

            if count == 1:
                subject = f"{verb} 1 file"
            else:
                subject = f"{verb} {count} files"
        else:
            # multiple operations, try comma-separated
            short_parts = [items[0][1]]
            for item in items[1:]:
                short_parts.append(item[1][0].lower() + item[1][1:])
            candidate = ", ".join(short_parts)
            if len(candidate) < GitRepo.SUBJECT_LINE_LIMIT:
                return candidate

            subject = f"Change {count} files"

        # couldn't fit in the subject line, itemise
        body_lines = [
            f"- {item[1]}"
                for item in items
        ]
        return f"{subject}\n\n" + "\n".join(body_lines)



from bolthole.debounce import collapse_events, Event


def test_single_created():
    assert collapse_events([
        Event("created", "a.txt"),
    ]) == [
        Event("created", "a.txt"),
    ]


def test_single_modified():
    assert collapse_events([
        Event("modified", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_single_deleted():
    assert collapse_events([
        Event("deleted", "a.txt"),
    ]) == [
        Event("deleted", "a.txt"),
    ]


def test_single_renamed():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_created_then_deleted():
    assert collapse_events([
        Event("created", "a.txt"),
        Event("deleted", "a.txt"),
    ]) == []


def test_created_then_modified():
    assert collapse_events([
        Event("created", "a.txt"),
        Event("modified", "a.txt"),
    ]) == [
        Event("created", "a.txt"),
    ]


def test_deleted_then_created():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("created", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_modified_then_deleted():
    assert collapse_events([
        Event("modified", "a.txt"),
        Event("deleted", "a.txt"),
    ]) == [
        Event("deleted", "a.txt"),
    ]


def test_modified_then_modified():
    assert collapse_events([
        Event("modified", "a.txt"),
        Event("modified", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_created_modified_deleted():
    assert collapse_events([
        Event("created", "a.txt"),
        Event("modified", "a.txt"),
        Event("deleted", "a.txt"),
    ]) == []


def test_deleted_created_deleted():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("created", "a.txt"),
        Event("deleted", "a.txt"),
    ]) == [
        Event("deleted", "a.txt"),
    ]


def test_created_modified_renamed():
    assert collapse_events([
        Event("created", "a.txt"),
        Event("modified", "a.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("created", "b.txt"),
    ]


def test_modified_renamed_deleted():
    assert collapse_events([
        Event("modified", "a.txt"),
        Event("renamed", "a.txt", "b.txt"),
        Event("deleted", "b.txt"),
    ]) == [
        Event("deleted", "a.txt"),
    ]


def test_rename_chain_two():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("renamed", "b.txt", "c.txt"),
    ]) == [
        Event("renamed", "a.txt", "c.txt"),
    ]


def test_rename_chain_three():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("renamed", "b.txt", "c.txt"),
        Event("renamed", "c.txt", "d.txt"),
    ]) == [
        Event("renamed", "a.txt", "d.txt"),
    ]


def test_rename_back_to_original():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("renamed", "b.txt", "a.txt"),
    ]) == []


def test_rename_cycle_three():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("renamed", "b.txt", "c.txt"),
        Event("renamed", "c.txt", "a.txt"),
    ]) == []


def test_three_way_cycle_simultaneous():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("renamed", "b.txt", "c.txt"),
        Event("renamed", "c.txt", "a.txt"),
    ]) == []


def test_created_then_renamed():
    assert collapse_events([
        Event("created", "a.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("created", "b.txt"),
    ]


def test_modified_then_renamed():
    assert collapse_events([
        Event("modified", "a.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_modified_modified_then_renamed():
    assert collapse_events([
        Event("modified", "a.txt"),
        Event("modified", "a.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_renamed_then_deleted():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("deleted", "b.txt"),
    ]) == [
        Event("deleted", "a.txt"),
    ]


def test_renamed_then_modified():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("modified", "b.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_renamed_then_created_at_source():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("created", "a.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
        Event("created", "a.txt"),
    ]


def test_renamed_then_modified_at_source():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("modified", "a.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_renamed_then_deleted_at_source():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("deleted", "a.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_two_creates():
    assert collapse_events([
        Event("created", "a.txt"),
        Event("created", "b.txt"),
    ]) == [
        Event("created", "a.txt"),
        Event("created", "b.txt"),
    ]


def test_modified_and_deleted():
    assert collapse_events([
        Event("modified", "a.txt"),
        Event("deleted", "b.txt"),
    ]) == [
        Event("modified", "a.txt"),
        Event("deleted", "b.txt"),
    ]


def test_renamed_and_created():
    assert collapse_events([
        Event("renamed", "a.txt", "b.txt"),
        Event("created", "c.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
        Event("created", "c.txt"),
    ]


def test_deleted_then_renamed_to_same():
    assert collapse_events([
        Event("deleted", "b.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("modified", "b.txt"),
    ]


def test_created_then_renamed_to_same():
    assert collapse_events([
        Event("created", "b.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
    ]


def test_modified_then_renamed_to_same():
    assert collapse_events([
        Event("modified", "b.txt"),
        Event("renamed", "a.txt", "b.txt"),
    ]) == [
        Event("deleted", "a.txt"),
        Event("modified", "b.txt"),
    ]


def test_three_way_swap():
    assert collapse_events([
        Event("renamed", "a.txt", "tmp.txt"),
        Event("renamed", "b.txt", "a.txt"),
        Event("renamed", "tmp.txt", "b.txt"),
    ]) == [
        Event("renamed", "a.txt", "b.txt"),
        Event("renamed", "b.txt", "a.txt"),
    ]


def test_atomic_replace_with_tmp_in_name():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("renamed", "tmp.txt", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_atomic_replace_with_swp_extension():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("renamed", "a.txt.swp", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_atomic_replace_with_backup_name():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("renamed", "backup.txt", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_atomic_replace_with_tilde_suffix():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("renamed", "a.txt~", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_atomic_replace_with_hash_prefix():
    assert collapse_events([
        Event("deleted", "a.txt"),
        Event("renamed", ".#a.txt", "a.txt"),
    ]) == [
        Event("modified", "a.txt"),
    ]


def test_atomic_replace_unrelated_filenames():
    assert collapse_events([
        Event("deleted", "config.json"),
        Event("renamed", "data.json", "config.json"),
    ]) == [
        Event("modified", "config.json"),
    ]



from dataclasses import dataclass


@dataclass(frozen=True)
class Event:
    type: str
    path: str
    new_path: str | None = None


def find_origin(
    renames: dict[str, str],
    path: str,
) -> str | None:
    for origin, dest in renames.items():
        if dest == path:
            return origin
    return None


def build_result(
    states: dict[str, str],
    renames: dict[str, str],
) -> list[Event]:
    result = []
    emitted = set()
    for origin, dest in sorted(renames.items()):
        if origin in emitted or dest == origin:
            continue
        if renames.get(dest) == origin:
            result.append(Event("renamed", origin, dest))
            result.append(Event("renamed", dest, origin))
            emitted.add(dest)
        else:
            result.append(Event("renamed", origin, dest))
        emitted.add(origin)

    for path, state in sorted(states.items()):
        result.append(Event(state, path))

    return result


def collapse_events(
    events: list[Event],
) -> list[Event]:
    states: dict[str, str] = {}
    renames: dict[str, str] = {}
    vacated: set[str] = set()

    for event in events:
        if event.type == "created":
            path = event.path
            if states.get(path) == "deleted":
                states[path] = "modified"
            else:
                states[path] = "created"
            vacated.discard(path)

        elif event.type == "modified":
            path = event.path
            if path in vacated:
                continue
            if find_origin(renames, path):
                continue
            if states.get(path) != "created":
                states[path] = "modified"

        elif event.type == "deleted":
            path = event.path
            if path in vacated:
                continue
            origin = find_origin(renames, path)
            if origin:
                del renames[origin]
                states[origin] = "deleted"
                continue
            if states.get(path) == "created":
                del states[path]
            else:
                states[path] = "deleted"

        elif event.type == "renamed":
            src, dst = event.path, event.new_path
            src_origin = find_origin(renames, src)
            if not src_origin:
                src_origin = src
            src_state = states.get(src)
            dst_state = states.get(dst)

            if dst_state == "deleted":
                states[dst] = "modified"
                if src_origin != src and src_origin in renames:
                    del renames[src_origin]
                vacated.add(src)
                continue

            if dst_state == "modified":
                states[src_origin] = "deleted"
                vacated.add(src)
                if src_origin in renames:
                    del renames[src_origin]
                continue

            if dst_state == "created":
                del states[dst]

            if src_state == "created":
                states.pop(src, None)
                states[dst] = "created"
            else:
                states.pop(src, None)
                renames[src_origin] = dst
                vacated.add(src)

    return build_result(states, renames)



import filecmp
import fnmatch
import os
import shutil
import signal
import stat
import threading
import time
from pathlib import Path
from types import FrameType

from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from bolthole.debounce import Event, collapse_events
from bolthole.git import GitRepo, output


def should_ignore(rel_path: str, patterns: list[str]) -> bool:
    parts = Path(rel_path).parts
    for pattern in patterns:
        if fnmatch.fnmatch(rel_path, pattern):
            return True
        for part in parts:
            if fnmatch.fnmatch(part, pattern):
                return True
    return False


def report_event(event: Event):
    if event.type == "renamed":
        output(f'   "{event.path}" renamed "{event.new_path}"')
    elif event.type == "modified":
        output(f'   "{event.path}" updated')
    else:
        output(f'   "{event.path}" {event.type}')


def report_action(event: Event):
    if event.type == "renamed":
        output(f'++ "{event.path}" -> "{event.new_path}"')
    elif event.type == "deleted":
        output(f'-- "{event.path}"')
    else:
        output(f'++ "{event.path}"')


def list_files(
    directory: Path,
    ignore_patterns: list[str],
) -> set[str]:
    files = set()
    for path in directory.rglob("*"):
        if path.is_file():
            rel = str(path.relative_to(directory))
            if not should_ignore(rel, ignore_patterns):
                files.add(rel)
    return files


def remove_empty_parents(
    path: Path,
    root: Path,
):
    parent = path.parent
    while parent != root:
        if not any(parent.iterdir()):
            parent.rmdir()
            parent = parent.parent
        else:
            break


def apply_event(
    event: Event,
    source: Path,
    dest: Path,
    dry_run: bool = False,
    verbose: bool = False,
):
    if verbose:
        report_event(event)
    report_action(event)

    if dry_run:
        if event.type in ("created", "modified"):
            output(f'#  copy "{event.path}"')
        elif event.type == "deleted":
            output(f'#  delete "{event.path}"')
        elif event.type == "renamed":
            output(f'#  rename "{event.path}" to "{event.new_path}"')
        return

    if event.type in ("created", "modified"):
        src = source / event.path
        dst = dest / event.path
        dst.parent.mkdir(parents=True, exist_ok=True)
        if dst.exists() and not os.access(dst, os.W_OK):
            os.chmod(dst, stat.S_IWUSR | stat.S_IRUSR)
        shutil.copy2(src, dst)
    elif event.type == "deleted":
        dst = dest / event.path
        dst.unlink(missing_ok=True)
        remove_empty_parents(dst, dest)
    elif event.type == "renamed":
        old_dst = dest / event.path
        new_dst = dest / event.new_path
        old_dst.unlink(missing_ok=True)
        remove_empty_parents(old_dst, dest)
        new_dst.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(source / event.new_path, new_dst)


def initial_sync(
    source: Path,
    dest: Path,
    ignore_patterns: list[str],
    dry_run: bool = False,
    show_git: bool = False,
    author: str | None = None,
    message: str | None = None,
    remotes: list[str] = [],
):
    if not dry_run:
        dest.mkdir(parents=True, exist_ok=True)

    source_files = list_files(source, ignore_patterns)
    if dest.exists():
        dest_files = list_files(dest, ignore_patterns)
    else:
        dest_files = set()

    events = []
    for rel_path in sorted(source_files):
        if rel_path not in dest_files:
            events.append(Event("created", rel_path))
        elif not filecmp.cmp(source / rel_path, dest / rel_path, shallow=False):
            events.append(Event("modified", rel_path))

    for rel_path in sorted(dest_files - source_files):
        events.append(Event("deleted", rel_path))

    for event in events:
        apply_event(event, source, dest, dry_run=dry_run, verbose=False)

    repo = GitRepo(
        dest,
        dry_run=dry_run,
        show_git=show_git,
        author=author,
        message=message,
    )
    committed = False
    if dry_run:
        if events:
            repo.commit_changes(events)
            committed = True
    else:
        uncommitted = repo.get_uncommitted()
        if uncommitted:
            repo.commit_changes(uncommitted)
            committed = True

    if remotes and committed:
        repo.push(remotes)


class DebouncingEventHandler(FileSystemEventHandler):
    def __init__(
        self,
        base_path: Path,
        ignore_patterns: list[str],
        dest_path: Path | None = None,
        debounce_delay: float = 0.1,
        dry_run: bool = False,
        verbose: bool = False,
        watchdog_debug: bool = False,
        show_git: bool = False,
        author: str | None = None,
        message: str | None = None,
        remotes: list[str] = [],
        grace: float = 0,
        bundle: float = 0,
    ):
        super().__init__()
        self.base_path = base_path
        self.dest_path = dest_path
        self.debounce_delay = debounce_delay
        self.dry_run = dry_run
        self.verbose = verbose
        self.watchdog_debug = watchdog_debug
        self.show_git = show_git
        self.author = author
        self.message = message
        self.remotes = remotes
        self.ignore_patterns = ignore_patterns
        self.grace = grace
        self.bundle = bundle
        self.pending_events: list[Event] = []
        self.lock = threading.Lock()
        self.timer: threading.Timer | None = None
        self.known_files = list_files(self.base_path, self.ignore_patterns)
        self.grace_events: dict[str, Event] = {}
        self.grace_timers: dict[str, threading.Timer] = {}
        self.grace_timestamps: dict[str, float] = {}
        self.commit_lock = threading.Lock()

    def relative_path(
        self,
        path: str,
    ) -> str:
        return str(Path(path).relative_to(self.base_path))

    def log_debug(
        self,
        event: Event,
    ):
        if not self.watchdog_debug:
            return
        if event.type == "renamed":
            output(f"watchdog: {event.type} {event.path} {event.new_path}")
        else:
            output(f"watchdog: {event.type} {event.path}")

    def queue_event(
        self,
        event: Event,
    ):
        self.log_debug(event)
        with self.lock:
            self.pending_events.append(event)
            if self.timer:
                self.timer.cancel()
            self.timer = threading.Timer(self.debounce_delay, self.flush_events)
            self.timer.start()

    def flush_events(
        self,
    ):
        with self.lock:
            if not self.pending_events:
                return
            events = self.pending_events[:]
            self.pending_events = []

        collapsed = collapse_events(events)
        for event in collapsed:
            if self.dest_path:
                apply_event(
                    event, self.base_path, self.dest_path,
                    dry_run=self.dry_run,
                    verbose=self.verbose,
                )
            elif self.verbose:
                report_event(event)

        if not collapsed:
            return

        if self.grace > 0:
            self.schedule_grace_commits(collapsed)
        else:
            if self.dest_path:
                repo_path = self.dest_path
            else:
                repo_path = self.base_path
            repo = GitRepo(
                repo_path,
                dry_run=self.dry_run,
                show_git=self.show_git,
                author=self.author,
                message=self.message,
            )
            repo.commit_changes(collapsed)
            if self.remotes:
                repo.push(self.remotes)

    def schedule_grace_commits(
        self,
        events: list[Event],
    ):
        now = time.time()
        with self.lock:
            for event in events:
                if event.type == "renamed":
                    path = event.new_path
                    old_path = event.path
                else:
                    path = event.path
                    old_path = None

                if old_path and old_path in self.grace_timers:
                    self.grace_timers[old_path].cancel()
                    del self.grace_timers[old_path]
                    old_event = self.grace_events.pop(old_path, None)
                    del self.grace_timestamps[old_path]
                    if old_event and old_event.type == "created":
                        # create + rename = create with new name
                        event = Event("created", path)

                if path in self.grace_timers:
                    self.grace_timers[path].cancel()
                    del self.grace_timers[path]
                    old_event = self.grace_events.get(path)
                    if old_event:
                        if old_event.type == "created" and event.type == "deleted":
                            # create + delete = nothing
                            del self.grace_events[path]
                            del self.grace_timestamps[path]
                            continue
                        elif old_event.type == "created":
                            # create + anything else = still created
                            event = Event("created", path)
                        elif old_event.type == "deleted" and event.type == "created":
                            # delete + create = modified
                            event = Event("modified", path)

                self.grace_events[path] = event
                self.grace_timestamps[path] = now
                timer = threading.Timer(
                    self.grace,
                    self.commit_after_grace,
                    args=[path],
                )
                self.grace_timers[path] = timer
                timer.start()

    def commit_after_grace(
        self,
        path: str,
    ):
        if self.bundle > 0:
            self.commit_bundled_files()
            return

        with self.lock:
            if path not in self.grace_events:
                return
            event = self.grace_events.pop(path)
            del self.grace_timers[path]
            del self.grace_timestamps[path]

        with self.commit_lock:
            if self.dest_path:
                repo_path = self.dest_path
            else:
                repo_path = self.base_path
            repo = GitRepo(
                repo_path,
                dry_run=self.dry_run,
                show_git=self.show_git,
                author=self.author,
                message=self.message,
            )
            repo.commit_changes([event])
            if self.remotes:
                repo.push(self.remotes)

    def commit_bundled_files(self):
        now = time.time()
        with self.lock:
            events_to_commit = []
            paths_to_remove = []
            for path, timestamp in self.grace_timestamps.items():
                age = now - timestamp
                if age >= self.grace:
                    events_to_commit.append(self.grace_events[path])
                    paths_to_remove.append(path)

            for path in paths_to_remove:
                del self.grace_events[path]
                del self.grace_timestamps[path]
                if path in self.grace_timers:
                    del self.grace_timers[path]

        if not events_to_commit:
            return

        with self.commit_lock:
            if self.dest_path:
                repo_path = self.dest_path
            else:
                repo_path = self.base_path
            repo = GitRepo(
                repo_path,
                dry_run=self.dry_run,
                show_git=self.show_git,
                author=self.author,
                message=self.message,
            )
            repo.commit_changes(events_to_commit)
            if self.remotes:
                repo.push(self.remotes)

    def on_created(
        self,
        event: FileSystemEvent,
    ):
        if event.is_directory:
            return
        path = self.relative_path(event.src_path)
        if should_ignore(path, self.ignore_patterns):
            return
        if path in self.known_files:
            self.queue_event(Event("modified", path))
        else:
            self.known_files.add(path)
            self.queue_event(Event("created", path))

    def on_modified(
        self,
        event: FileSystemEvent,
    ):
        if event.is_directory:
            return
        path = self.relative_path(event.src_path)
        if should_ignore(path, self.ignore_patterns):
            return
        self.queue_event(Event("modified", path))

    def on_deleted(
        self,
        event: FileSystemEvent,
    ):
        if event.is_directory:
            return
        path = self.relative_path(event.src_path)
        if should_ignore(path, self.ignore_patterns):
            return
        self.known_files.discard(path)
        self.queue_event(Event("deleted", path))

    def on_moved(
        self,
        event: FileSystemEvent,
    ):
        if event.is_directory:
            return
        src_path = self.relative_path(event.src_path)
        dst_path = self.relative_path(event.dest_path)
        src_ignored = should_ignore(src_path, self.ignore_patterns)
        dst_ignored = should_ignore(dst_path, self.ignore_patterns)

        if src_ignored and dst_ignored:
            return
        elif src_ignored:
            self.known_files.add(dst_path)
            self.queue_event(Event("created", dst_path))
        elif dst_ignored:
            self.known_files.discard(src_path)
            self.queue_event(Event("deleted", src_path))
        else:
            self.known_files.discard(src_path)
            self.known_files.add(dst_path)
            self.queue_event(Event("renamed", src_path, dst_path))


def watch(
    source: Path,
    dest: Path | None = None,
    dry_run: bool = False,
    verbose: bool = False,
    watchdog_debug: bool = False,
    ignore_patterns: list[str] = [],
    show_git: bool = False,
    source_label: str | None = None,
    dest_label: str | None = None,
    once: bool = False,
    author: str | None = None,
    message: str | None = None,
    remotes: list[str] = [],
    grace: float = 0,
    bundle: float = 0,
):
    ignore_patterns = [".git", ".gitignore"] + ignore_patterns

    if dest:
        initial_sync(
            source, dest,
            dry_run=dry_run,
            ignore_patterns=ignore_patterns,
            show_git=show_git,
            author=author,
            message=message,
            remotes=remotes,
        )
    else:
        repo = GitRepo(
            source,
            dry_run=dry_run,
            show_git=show_git,
            author=author,
            message=message,
        )
        events = repo.get_uncommitted()
        if events:
            repo.commit_changes(events)
            if remotes:
                repo.push(remotes)

    if once:
        return

    handler = DebouncingEventHandler(
        source,
        dest_path=dest,
        dry_run=dry_run,
        verbose=verbose,
        watchdog_debug=watchdog_debug,
        ignore_patterns=ignore_patterns,
        show_git=show_git,
        author=author,
        message=message,
        remotes=remotes,
        grace=grace,
        bundle=bundle,
    )
    observer = Observer()
    observer.schedule(handler, str(source), recursive=True)

    def sigterm_handler(
        signum: int,
        frame: FrameType | None,
    ):
        raise KeyboardInterrupt

    signal.signal(signal.SIGTERM, sigterm_handler)

    observer.start()

    if dest_label:
        output(f"   Copying {source_label} to {dest_label}...")
    else:
        output(f"   Watching {source_label}...")

    try:
        while True:
            observer.join(timeout=1)
            if not observer.is_alive():
                break
    except KeyboardInterrupt:
        pass
    finally:
        observer.stop()
        observer.join()
        handler.flush_events()
